# -*- coding: utf-8 -*-
"""
Utils
==========================

Utility functions for the Distance Closure package
"""
#    Copyright (C) 2015 by
#    Luis Rocha <rocha@indiana.edu>
#    Thiago Simas <@.>
#    Rion Brattig Correia <rionbr@gmail.com>
#    All rights reserved.
#    MIT license.
import numpy as np

__author__ = """\n""".join([
	'Luis Rocha <rocha@indiana.com>',
	'Thiago Simas <@.>',
	'Rion Brattig Correia <rionbr@gmail.com>',
	])

__all__ = [
			'dist2prox',
			'prox2dist',
		]

def prox2dist(P):
	"""
	Transforms a matrix of non-negative ``[0,1]`` proximities P to distance weights in the ``[0,inf]`` interval:
	
	.. math::

		d = \\frac{1}{p} - 1
	
	Args:
		P (matrix): Proximity matrix
	
	Returns:
		D (matrix): Distance matrix

	See Also:
		:attr:`dist2prox`
	"""
	if (type(P).__module__ == 'numpy'):
		return _prox2dist_numpy(P)
	elif (type(P).__module__.split('.')[1] == 'sparse'):
		return _prox2dist_sparse(P)
	else:
		raise ("Format not accepted: try numpy or scipy.sparse formats")


def _prox2dist_sparse(A):
	A.data = prox2dist_numpy(A.data)
	return A

def _prox2dist_numpy(A):
	
	def _prox2dist(x):
		if x == 0:
			return np.inf
		else:
			return (1/float(x)) - 1

	f = np.vectorize(_prox2dist)
	return f(A)

def dist2prox(D):
	"""
	Transforms a matrix of non-negative integer distances ``D`` to proximity/similarity weights in the ``[0,1]`` interval:
	
	.. math::
	
		p = \\frac{1}{(d+1)}

	It accepts both dense and sparse matrices.

	Args:
		D (matrix): Distance matrix

	Returns:
		P (matrix): Proximity matrix

	See Also:
		:attr:`prox2dist`

	"""
	if (type(D).__module__ == 'numpy'):
		return _dist2prox_numpy(D)
	elif (type(D).__module__.split('.')[1] == 'sparse'):
		return _dist2prox_numpy(D)
	else:
		raise ("Format not accepted: try numpy or scipy.sparse formats")

def _dist2prox_sparse(A):
	A.data = dist2prox_numpy(A.data)
	return A

def _dist2prox_numpy(A):

	def _dist2prox(x):
		if x == np.inf:
			return 0
		else:
			return (x + 1) ** -1

	f = np.vectorize(_dist2prox)
	return f(A)





# -*- coding: utf-8 -*-
"""
Transitive Closure
===================

Computes transitive closure on a graph Adjacency Matrix.

These algorithms work with undirected weighted (distance) graphs.
"""
#    Copyright (C) 2015 by
#    Luis Rocha <rocha@indiana.edu>
#    Thiago Simas <@.>
#    Rion Brattig Correia <rionbr@gmail.com>
#    All rights reserved.
#    MIT license.
import numpy as np
__name__ = 'distanceclosure'
__author__ = """\n""".join(['Luis Rocha <rocha@indiana.com>',
							'Thiago Simas <@.>',
							'Rion Brattig Correia <rionbr@gmail.com>'])
__all__ = ['transitive_closure']
__metrics__ = ['metric','ultrametric']
#
#
#
def transitive_closure(D, kind='metric', *args, **kwargs):
	"""
	Compute the transitive closure (All-Pairs-Shortest-Paths; APSP) using different shortest path measures
	on the distance graph (adjacency matrix) with values in the ``[0,inf]`` interval.
	
	.. math::
		c_{ij} = min_{k}( metric ( a_{ik} , b_{kj} ) )

	Args:
		D (matrix): A numpy dense matrix.
		kind (string): type of closure to compute: ``metric`` or ``ultrametric``.
		verbose (bool): Prints statements as it computes.

	Returns:
		C (matrix): transitive closure dense matrix

	Warnings:
		This algorithm uses dense matrix calculations which is still slow for very large graphs.
	
	Examples:

		>>> P = np.array([
			[1.,.9,.1,0.],
			[.9,1.,.8,0.],
			[.1,.8,1.,.6],
			[0.,0.,.6,1.],
			], dtype=float)
		>>> D = prox2dist(P)
		
		>>> transitive_closure(D, kind='metric', verbose=True)
			[ 0. ,.11111111, 0.36111111, 1.02777778],
			[ 0.11111111, 0., 0.25, 0.91666667],
			[ 0.36111111, 0.25, 0., 0.66666667],
			[ 1.02777778, 0.91666667, 0.66666667,  0.]

	Note:
		Metric: :math:`(min,+)`
		
		Ultrametric: :math:`(min,max)` -- also known was maximum flow.
		
		Semantic proximity: TODO
		
		.. math::
			
				[ 1 + \\sum_{i=2}^{n-1} log k(v_i) ]^{-1}


	"""
	_check_for_metric_type(kind)
	
	# Numpy object
	if (type(D).__module__ == np.__name__):
		return _transitive_closure_numpy(D, kind, *args, **kwargs)
	else:
		raise TypeError("Input is not a numpy object")

def _transitive_closure_numpy(A, kind='metric', verbose=False):
	"""
	Calculates Transitive Closure ising numpy dense matrix traversing.
	"""
	C = A.copy()
	n,m = A.shape
	
	# Check if diagonal is all zero
	if sum( np.diagonal(A) ) > 0:
		raise ValueError("Diagonal has to be zero for matrix computation to be correct")

	# Compute Transitive Closure
	for i in xrange(0,n):
		if verbose:
			print 'calc row:',i+1,'of',m
		for j in xrange(0,n):
			
			if kind == 'metric':
				vec = C[i,:] + C[:,j]
				C[i,j] = vec.min()

			elif kind == 'ultrametric':
				vec = np.maximum( C[i,:], C[:,j] )
				C[i,j] = vec.min()
	
	return np.array(C)


def _transitive_closure_dijkstra_matrix_all_pairs(A, kind):
	"""
	TODO
	"""
	# TODO	
	return None
	n,m = A.shape

	for source in xrange(0,n):
		_transitive_closure_dijkstra_matrix(A, source, kind)

def _transitive_closure_dijkstra_matrix(A, source, kind):
	"""
	TODO
	"""
	# TODO
	pass

def _check_for_metric_type(kind):
	"""
	Check for available metric functions.
	"""
	if kind not in __metrics__:
		raise TypeError("Metric not found for this algorithm. Try 'metric' or 'ultrametric',")




# -*- coding: utf-8 -*-
"""
Backbone Extraction
===================

Extracts the Backbone edges from the original graph and the distance closure computation.
"""
#    Copyright (C) 2015 by
#    Luis Rocha <rocha@indiana.edu>
#    Thiago Simas <@.>
#    Rion Brattig Correia <rionbr@gmail.com>
#    All rights reserved.
#    MIT license.
import numpy as np
__name__ = 'distanceclosure'
__author__ = """\n""".join(['Luis Rocha <rocha@indiana.com>',
							'Thiago Simas <@.>',
							'Rion Brattig Correia <rionbr@gmail.com>'])
__all__ = ['backbone']
#
#
#
def backbone(A, C):
	"""
	Return backbone edges based on the the original graph and its transitive closure.
	By definition, the backbone are the edges that did not change value in the transitive closure computation.

	Args:
		A (matrix): Adjacency matrix from original graph.
		C (matrix): Adjacency matrix from transitive closure graph

	Returns:
		B (array): Adjacency matrix where backbone edges are ``1``, semi-metric edges are ``2``, diagonal is ``-1``, ``0`` otherwise.

	Note:
		Inputs accepts both dense and sparse matrices.

	Examples:
		>>> C = transitive_closure_numpy(A, kind='metric')
		>>> B = backbone_numpy(A, C)
		
		You can then access the backbone edges by using:
		
		>>> import numpy as np
		>>> rows, cols = np.where(B==1)
	"""	
	# Check for data type
	if (type(A).__module__ == np.__name__) and (type(C).__module__ == np.__name__):
		return _backbone_numpy(A, C)
	else:
		raise TypeError("Inputs are not valid objects, try numpy or scipy objects. Objects must be both the same type.")

def _backbone_numpy(A, C):
	n,m = A.shape
	B = np.zeros(A.shape)
	if n==m:
		np.fill_diagonal(B, -1)
	# Semi-metric values = 2
	rows, cols = np.where( (C > 0) & (C != np.inf) )
	B[rows,cols] = 2
	# Metric Values = 1
	rows, cols = np.where( (C > 0) & (C != np.inf) & np.isclose(A , C) )
	B[rows,cols] = 1
	return B

